-- By JugadorXEI#1815 on Discord, @JugadorXEI on Twitter.
--[[
	WHAT IS ITS PURPOSE:
		Context Tags allows to check if a number of console variables are meeting a specific value,
		and if so, enable a number of tags inside of a map.
		For example, to check if a map is running an acrobatics mod, one would check the cvars
		ac_active or as_active, and see if this value is 1.
		If so, the script will check if there's linedef executor triggers with the tags defined.
		If this is the case, they will be executed like a regular linedef executor.

	HOW DOES IT WORK:
		There's four types of Lua level headers to look for:
			Lua.ContextTags_CVarsToCheck#:
				Defines which console variables (cvars) to check for.
				More than one can be checked with a comma-delimited list.
				Input example: kartspeed,slipstream_enabled,ac_active 

			Lua.ContextTags_ValueToMeet#:
				Optional but should not be skipped for clarity's sake.
				Defines which values should the cvars be compared to.
				You are strongly suggested to use numbers instead of strings for comparison purposes.
				More than one value can be checked for with a comma-delimited list.
				Input example: hard,1,0

				Assuming there's more cvars than values:
				If no values are set:
					1 is assumed as every value that each cvar should have.
				If at least (more than) one is set:
					The first value will be assumed to be met for every other cvar.

			Lua.ContextTags_AllCvarsTrue#:
				Optional. Defines if every value defined for each cvar must be true.
				Otherwise, at least one value must be true for any cvar for executors to trigger.
				The script only checks if this header is defined, and will apply if it does.
				Input example: any cute text you wish for it to be here, could also be a comment too.

			Lua.ContextTags_ExecuteTags#:
				Defines which tags to execute if the cvar(s) had their value(s) met.
				More than one can be executed with a comma-delimited list.
				Input example: 5,6,7,8,9

			The # in each level header must be a number. For example: Lua.ContextTags_CVarsToCheck1
			You can have multiple checks doing different things. An example of multiple checks would be:

			Lua.ContextTags_CVarsToCheck1 = ac_active,as_active
			Lua.ContextTags_ExecuteTags1 = 1,2,3,4
			Lua.ContextTags_CVarsToCheck2 = kartspeed,slipstream_enabled,ac_active
			Lua.ContextTags_ValueToMeet2 = hard,1,0
			Lua.ContextTags_AllCvarsTrue2 = sure, whatever
			Lua.ContextTags_ExecuteTags2 = 5,6,7,8,9

			The level header check is done programatically - 
			the script will check if ContextTags_CVarsToCheck1 exists, and if not,
			it will skip reading the next level headers, even if ContextTags_CVarsToCheck2 and
			ContextTags_CVarsToCheck3 were to exist. Make sure not to have any typos on your level header.

			If you are not too sure if something is working right, set the debug variable to true
			inside the script.

	WHY:
		There's various reasons to use Context Tags:
			- Balancing for mods is desired while keeping the vanilla experience intact.
				Mods such as slipstream thrive when there's sneaker panels to stack with,
				and acrobatics mods thrive on the presence of springs and ramps.
				However, overusing these elements can worsen the general vanilla experience,
				specially if these elements are placed haphazardly for the purpose of these mods.
				Thus, this mod provices the best of both worlds in these instances, allowing
				a greater amount of control over what the level looks and behaves in modded
				servers.

			- A gimmick that works in vanilla breaks or is worsened when a mod is running.
				If this issue is known, the level designer can check if the mod is running
				and circumvent such issues or even improve it for modded gameplay.

			- Easter eggs.
				It can interesting to allow easter eggs or secret paths if certain conditions
				inside the server are met.
]]

local CVARHEADERLINEDEFVERSION = 3

if jug_headerlinedef == nil then
	rawset(_G, "jug_headerlinedef", {})
end

if jug_headerlinedef.ver == nil or jug_headerlinedef.ver < CVARHEADERLINEDEFVERSION then

if jug_headerlinedef.ver == nil then
jug_headerlinedef.contexttags_force = CV_RegisterVar({"ctags_force", "", CV_NETVAR})
end

-- Enable debug here:
local debug = true -- DISABLE THIS ON RELEASES
local function debugprint(string)
	if debug then print(string) end
end

local function turnCommaDelimitedStringIntoTable(string)
	if string == nil or string == "" then
		return {}, 0
	end

	local tableVar = {}
	local newString = ""
	local numThingsAdded = 0
	for i = 1, #string do
		local c = string:sub(i,i)
		
		if c == ',' then
			table.insert(tableVar, newString)
			numThingsAdded = $ + 1
			newString = ""
		else
			newString = $..c
		end
	end

	if newString ~= "" then
		table.insert(tableVar, newString);
		numThingsAdded = $ + 1
	end

	return tableVar, numThingsAdded
end

jug_headerlinedef.tagLoaderHandler = function()
	if leveltime != 2 then return end

	-- the first one should at least exist
	if mapheaderinfo[gamemap].contexttags_cvarstocheck1 == nil then return end;
	
	-- how many to check?
	local CVarHeadersToCheck = 0;
	while true
		if not mapheaderinfo[gamemap]["contexttags_cvarstocheck" + (CVarHeadersToCheck + 1)] then
			break
		end

		CVarHeadersToCheck = $ + 1;
	end

	debugprint("Headers to check: " + CVarHeadersToCheck);

	-- are there any cvars to force?
	local forcingCVars = false
	local CVarsToForce = jug_headerlinedef.contexttags_force.string
	local CVarsToForceTable = {}
	if CVarsToForce ~= "" then
		CVarsToForceTable = turnCommaDelimitedStringIntoTable(CVarsToForce)
		forcingCVars = true
	end

	for i = 1, CVarHeadersToCheck do
		-- Check if each var meets the value.
		local CVars = mapheaderinfo[gamemap]["contexttags_cvarstocheck" + i]
		local separatedCVars, separatedCVarsNum = turnCommaDelimitedStringIntoTable(CVars)

		-- are we forcing this cvar?
		local shouldAllCvarsBeTrue = (mapheaderinfo[gamemap]["contexttags_allcvarstrue" + i] ~= nil)

		local forced = false
		if forcingCVars then
			local numOfCvarsForced = 0			

			for key, valueCvarForce in pairs(CVarsToForceTable) do
				for key, valueCVar in pairs(separatedCVars) do
					if valueCvarForce == valueCVar then
						if not shouldAllCvarsBeTrue then
							forced = true
							break
						else
							numOfCvarsForced = $ + 1
							continue
						end
						
					end
				end

				if forced and (not shouldAllCvarsBeTrue) then break end
			end

			debugprint("how many cvars to force " + separatedCVarsNum)
			if shouldAllCvarsBeTrue and numOfCvarsForced == separatedCVarsNum then
				debugprint("ok we met every cvar ever: " + separatedCVarsNum)
				forced = true
			end

			if forced then
				debugprint("forcing cvar " + valueCVar + "...")
			end
		end

		if not forced then 
			local values = mapheaderinfo[gamemap]["contexttags_valuetomeet" + i]
			local valuesTable, numValues = turnCommaDelimitedStringIntoTable(values)

			-- are we dealing with a number or string? Used to know what to compare from the cvar object.
			if numValues > 0 then
				for i = 1, numValues
					local thisValue = tonumber(valuesTable[i])
					if type(thisValue) == "number" then
						-- debugprint("IT'S A NUMBER!!")
						valuesTable[i] = thisValue
					end
				end
			end

			-- do we have less values than cvars?
			-- add 1 as a value if none is set.
			-- we'll assume the user wants it enabled.
			-- add the first value in the list if one or more is set.
			if numValues < separatedCVarsNum then
				for i = numValues + 1, separatedCVarsNum do
					if numValues == 0 then
						table.insert(valuesTable, 1)
					else
						table.insert(valuesTable, valuesTable[1])
					end

					if debug then numValues = $ + 1 end
				end
			end

			if debug then
				values = ""
				
				for i = 1, numValues do
					values = $ + tostring(valuesTable[i])

					if i ~= numValues then
						values = $ + ","
					end
				end 
			end

			-- time to compare these values with the cvars, if they exist.
			local shouldExecuteTags = shouldAllCvarsBeTrue -- this makes sense, trust me.
			debugprint("allcvarsbetrue " + shouldAllCvarsBeTrue)
			debugprint("shouldexecutetags " + shouldExecuteTags)

			for i = 1, separatedCVarsNum do
				if shouldAllCvarsBeTrue then
					shouldExecuteTags = false
				end

				local thisCVar = CV_FindVar(separatedCVars[i])
				if thisCVar == nil then continue end

				local valueToCompareTo = valuesTable[i]
				local isANumber = type(valueToCompareTo) == "number"

				local CVarValue;
				if isANumber then 	CVarValue = thisCVar.value
				else				CVarValue = thisCVar.string end

				-- debugprint("isnumber: " + isANumber + '\n' + CVarValue + " = " + valueToCompareTo)

				if (not isANumber and CVarValue:lower() == valueToCompareTo:lower()) or
					(CVarValue == valueToCompareTo) then

					shouldExecuteTags = true

					if shouldAllCvarsBeTrue then continue
					else break end

				elseif shouldAllCvarsBeTrue then break
				end
			end

			if not shouldExecuteTags then
				debugprint("The cvars \"" + CVars +  "\" did not match the value " + values + ", next.")
				continue
			else
				debugprint("CVar " + CVars + " has value " + values + " - executing tags...")
			end
		end

		-- ok, but how many tags do we need to execute?
		local tagsHeaderString = mapheaderinfo[gamemap]["contexttags_executetags" + i]
		local tags = turnCommaDelimitedStringIntoTable(tagsHeaderString);

		-- ok, let's execute all of the tags in the table.
		for key, value in pairs(tags) do
			local number = tonumber(value)
			debugprint(CVars + ": executing tag " + number)
			P_LinedefExecute(number)
		end
	end

	debugprint("done")
end

if jug_headerlinedef.ver == nil then
	addHook("ThinkFrame", function()
		jug_headerlinedef.tagLoaderHandler();
	end)
end

-- end
jug_headerlinedef.ver = CVARHEADERLINEDEFVERSION
end

-- By JugadorXEI#1815 on Discord, @JugadorXEI on Twitter.